import os
import base64
import json

import numpy as np
# import faiss
from google.cloud import storage
from google.cloud import bigtable
from google.cloud.bigtable import row_filters, row_set

from .query_result import IdsQR, EmptyQR, IdsScoresQR
from .analyzers import analyze
from .row_parsers import row_key_formatters, row_parsers
from .query_validator import getTypeOfFieldFromMapping


from cachetools import cached, LRUCache, TTLCache

def download_gcs_file(project_id, gcs_uri, file_name):
    assert gcs_uri.startswith('gs://')
    a = gcs_uri[5:]
    storage_client = storage.Client(project_id)
    bucket_name = a.split('/')[0]
    blob_name = '/'.join(a.split('/')[1:])
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(blob_name)
    blob.download_to_filename(file_name)

def download_gcs_dir(project_id=None, bucket_name=None, prefix=None, dl_dir=None):
    storage_client = storage.Client(project_id)
    bucket = storage_client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=prefix)  # Get list of files
    for blob in blobs:
        filename = blob.name.replace(prefix, '') 
        print('downloading %s' % (filename))
        blob.download_to_filename(dl_dir + filename)

class Index(object):
    """
    The data. Contains inverted index postings, faiss indices.
    """
    def __init__(self, name, 
        project_id='',
        bt_instance_id='',
        bt_table_id='',
        schema_uri=None,
        schema=None):

        self.name = name
        self.project_id = project_id
        self.init_data_dir()

        assert schema_uri is not None or schema is not None
        if schema is not None:
            self.schema = schema
        else:
            self.schema_uri = schema_uri
            self.schema = self.load_schema()
        self.mapping = self.schema.get('mapping', {})

        self.bt_instance_id = bt_instance_id
        self.bt_table_id = bt_table_id
        
        self.bt_client = bigtable.Client(project=self.project_id)
        self.bt_instance = self.bt_client.instance(self.bt_instance_id)
        self.bt_table = self.bt_instance.table(self.bt_table_id)
        self.bt_row_filter = row_filters.CellsColumnLimitFilter(1)

        self.ttl_for_ttled_postings = int(self.schema.get('ttl', 60*5)) # 5 minutes default
        self.ttled_fields = set(self.schema.get('ttled_fields', []))
        
        self.lru_cache = LRUCache(10000)
        self.ttl_cache = TTLCache(10000, self.ttl_for_ttled_postings)
    
    def init_data_dir(self):
        if not os.path.exists('data'):
            os.mkdir('data')
        if not os.path.exists(os.path.join('data', self.name)):
            os.mkdir(os.path.join('data', self.name))

    def load_schema(self):
        schema_filename = os.path.join('data', self.name, 'schema.json')
        download_gcs_file(self.project_id, self.schema_uri, schema_filename)
        
        schema_str = ''
        with open(schema_filename, 'r') as f:
            schema_str = ' '.join(f.readlines())
        
        schema = json.loads(schema_str)
        return schema
    
    def btLookup(self, row_key):
        row = self.bt_table.read_row(row_key.encode('utf-8'), self.bt_row_filter)
        return row


    def getPostings(self, field, value):
        field_type = getTypeOfFieldFromMapping(field, self.mapping)
        row_key = row_key_formatters[field_type](field, value)

        cached_value = None

        has_ttl = field in self.ttled_fields

        if has_ttl:
            cached_value = self.ttl_cache.get(row_key)
        else:
            cached_value = self.lru_cache.get(row_key)
        
        if cached_value is not None:
            return cached_value
        else:
            row = self.btLookup(row_key)
            fetched_qr = None
            if row is not None:
                fetched_qr = row_parsers[field_type](row)
            else:
                fetched_qr = EmptyQR()
            if has_ttl:
                self.ttl_cache[row_key] = fetched_qr
            else:
                self.lru_cache[row_key] = fetched_qr
            return fetched_qr

    def has_row_key_cached(self, row_key, field=None):
        if field is not None:
            if field in self.ttled_fields:
                return row_key in self.ttl_cache
            else:
                return row_key in self.lru_cache
        else:
            return row_key in self.lru_cache or row_key in self.ttl_cache


    def getDocument(self, i):
        row_key = 'doc|%s' % (i)
        row = self.btLookup(row_key)
        doc = None
        if row is not None:
            doc = row_parsers['doc'](row)
        return doc
    
    def getDocuments(self, doc_ids):
        row_keys = ['doc|%s' % (i) for i in doc_ids]

        documents = []

        rs = row_set.RowSet()
        set_size = 0
        for rk in row_keys:
            rs.add_row_key(rk)
            set_size += 1
        if set_size > 0:
            partial_rows_data = self.bt_table.read_rows(row_set=rs, filter_=self.bt_row_filter)
            partial_rows_data.consume_all()

            for rk in row_keys:
                row = partial_rows_data.rows.get(rk.encode())
                if row is not None:
                    doc = row_parsers['doc'](row)
                    documents.append(doc)
                else:
                    documents.append(None)
        return documents
    
    def analyze(self, field, value, return_text=True):
        spec = self.mapping
        for f in field.split('.'):
            spec = spec.get('properties', spec)
            spec = spec.get(f, {})
        analyzer_type = spec.get('analyzer', 'english')
        return analyze(value, analyzer=analyzer_type, return_text=return_text)


import base64
from google.cloud.bigtable import row_set

from .query_result import EmptyQR

from .row_parsers import row_parsers, row_key_formatters

from .query_validator import getTypeOfFieldFromMapping

from .analyzers import analyze_from_mapping


# ************ assume query has been validated ************ #



def boolTypesGetter(query, mapping):
    rkt = {}
    for key in query:
        rkt.update(row_key_type_getters[key](query[key], mapping))
    return rkt

def boolChildTypesGetter(query, mapping):
    rkt = {}
    for item in query:
        op = [k for k in item][0]
        rkt.update(row_key_type_getters[op](item[op], mapping))
    return rkt

def termTypesGetter(query, mapping):
    rkt = {}
    field = [k for k in query][0]
    value = query[field]['value']
    row_key = row_key_formatters['keyword'](field, value)
    rkt[row_key] = {
        'field': field,
        'type': 'keyword'
    }
    return rkt

def termsTypesGetter(query, mapping):
    rkt = {}
    field = [k for k in query][0]
    
    for value in query[field]:
        row_key = row_key_formatters['keyword'](field, value)
        rkt[row_key] = {
            'field': field,
            'type': 'keyword'
        }
    return rkt

def matchTypesGetter(query, mapping):
    rkt = {}
    field = [k for k in query][0]
    match_string = query[field]
    tokens = analyze_from_mapping(field, match_string, mapping, return_text=False)

    for value in tokens:
        row_key = row_key_formatters['keyword'](field, value)
        rkt[row_key] = {
            'field': field,
            'type': 'keyword'
        }
    return rkt

def rangeTypesGetter(query, mapping):
    rkt = {}
    field = [k for k in query][0]
    field_type = getTypeOfFieldFromMapping(field, mapping)
    row_key = row_key_formatters[field_type](field, None)
    rkt[row_key] = {
        'field': field,
        'type': field_type
    }
    return rkt

def fieldBoostTypesGetter(query, mapping):
    rkt = {}
    field = query['field']
    field_type = getTypeOfFieldFromMapping(field, mapping)
    row_key = row_key_formatters[field_type](field, None)
    rkt[row_key] = {
        'field': field,
        'type': field_type
    }
    return rkt

row_key_type_getters = {
    'bool': boolTypesGetter,
    'filter': boolChildTypesGetter,
    'must': boolChildTypesGetter,
    'should': boolChildTypesGetter,
    'must_not': boolChildTypesGetter,
    'term': termTypesGetter,
    'terms': termsTypesGetter,
    'match': matchTypesGetter,
    'range': rangeTypesGetter,
    'fieldBoost': fieldBoostTypesGetter,
}

# given a query and an index mapping
# return a dictionary containing all the rowkeys
# the query needs and their types
def getRowKeysAndTypes(query, mapping):
    op = [k for k in query][0]
    row_keys_types = row_key_type_getters[op](query[op], mapping)
    return row_keys_types


# given a query and an index,
# efficiently fetch all the postings the query will need
# returns a dictionary of row_key to QueryResult
def prefetchPostings(query, index):
    row_keys_types = getRowKeysAndTypes(query, index.mapping)

    rs = row_set.RowSet()
    set_size = 0
    for rk in row_keys_types:
        if not index.has_row_key_cached(rk, field=row_keys_types[rk]['field']):
            rs.add_row_key(rk)
            set_size += 1
    if set_size > 0:
        partial_rows_data = index.bt_table.read_rows(row_set=rs, filter_=index.bt_row_filter)
        partial_rows_data.consume_all()

        # rowkey_to_qr = {}

        for rk in row_keys_types:
            row = partial_rows_data.rows.get(rk.encode())
            t = row_keys_types[rk]['type']
            f = row_keys_types[rk]['field']
            qr = None
            if row is None:
                # rowkey_to_qr[rk] = EmptyQR()
                qr = EmptyQR()
            else:
                # rowkey_to_qr[rk] = row_parsers[t](row)
                qr = row_parsers[t](row)
            
            if f in index.ttled_fields:
                # index.ttl_cache[rk] = rowkey_to_qr[rk]
                index.ttl_cache[rk] = qr
            else:
                # index.lru_cache[rk] = rowkey_to_qr[rk]
                index.lru_cache[rk] = qr
        
        # return rowkey_to_qr


import base64
import json

import numpy as np

from .query_result import IdsQR, IdsScoresQR

def keyword_row_parser(row):
    ids_bytes = row.cells.get('i').get('i'.encode())[0].value
    ids = np.array(np.frombuffer(ids_bytes, dtype=np.int32), dtype=np.int32)
    return IdsQR().fromIds(ids)

def integer_row_parser(row):
    ids_bytes = row.cells.get('i').get('i'.encode())[0].value
    ids = np.array(np.frombuffer(ids_bytes, dtype=np.int32), dtype=np.int32)
    scores_bytes = row.cells.get('s').get('i'.encode())[0].value
    scores = np.array(np.frombuffer(scores_bytes, dtype=np.float32), dtype=np.float32)
    return IdsScoresQR(ids, scores)

def float_row_parser(row):
    ids_bytes = row.cells.get('i').get('i'.encode())[0].value
    ids = np.array(np.frombuffer(ids_bytes, dtype=np.int32), dtype=np.int32)
    scores_bytes = row.cells.get('s').get('f'.encode())[0].value
    scores = np.array(np.frombuffer(scores_bytes, dtype=np.float32), dtype=np.float32)
    return IdsScoresQR(ids, scores)

def doc_row_parser(row):
    response = row.cells.get('doc').get('doc'.encode())[0].value
    try:
        return json.loads(response)
    except:
        return response

row_parsers = {
    'keyword': keyword_row_parser,
    'text': keyword_row_parser,
    'integer': integer_row_parser,
    'float': float_row_parser,
    'doc': doc_row_parser
}


def keyword_row_key_formatter(field, value):
    encoded_value = base64.b64encode(value.encode()).decode('utf-8')
    row_key = '%s|%s' % (field, encoded_value)
    return row_key

def integer_row_key_formatter(field, value):
    return field

def float_row_key_formatter(field, value):
    return field

def boolean_row_key_formatter(field, value):
    if type(value) is str:
        return '%s|%s' % (field, value)
    else:
        return '%s|%s' % (field, 'true' if value else 'false')

def doc_row_key_formatter(doc_id, _):
    return 'doc|%s' % (doc_id)

row_key_formatters = {
    'keyword': keyword_row_key_formatter,
    'text': keyword_row_key_formatter,
    'integer': integer_row_key_formatter,
    'float': float_row_key_formatter,
    'boolean': boolean_row_key_formatter,
    'doc': doc_row_key_formatter
}


numeric_field_types = {'integer', 'float'}
keyword_like_types = {'keyword', 'text', 'boolean'}

def getTypeOfFieldFromMapping(field, mapping):
    assert type(field) is str
    field_type = None
    spec = {}
    spec.update(mapping)
    for f in field.split('.'):
        spec = spec.get('properties', spec)
        spec = spec.get(f, {})
    field_type = spec.get('type')
    return field_type


def validateBool(query, mapping):
    assert type(query) is dict
    allowed_keys = {'must', 'should', 'filter', 'must_not'}
    for key in query:
        assert key in allowed_keys
        query_validators[key](query[key], mapping)
    return True

def validateFilter(query, mapping):
    assert type(query) is list
    for item in query:
        assert type(item) is dict
        assert len(item) == 1
        key = [k for k in item][0]
        query_validators[key](item[key], mapping)
    return True

def validateMustNot(query, mapping):
    assert type(query) is list
    for item in query:
        assert type(item) is dict
        assert len(item) == 1
        key = [k for k in item][0]
        query_validators[key](item[key], mapping)
    return True

def validateMust(query, mapping):
    assert type(query) is list
    for item in query:
        assert type(item) is dict
        assert len(item) == 1
        key = [k for k in item][0]
        query_validators[key](item[key], mapping)
    return True

def validateShould(query, mapping):
    assert type(query) is list
    for item in query:
        assert type(item) is dict
        assert len(item) == 1
        key = [k for k in item][0]
        query_validators[key](item[key], mapping)
    return True

def validateTerm(query, mapping):
    assert type(query) is dict
    assert len(query) == 1

    field = [k for k in query][0]
    field_type = getTypeOfFieldFromMapping(field, mapping)
    assert field_type is not None
    assert field_type in keyword_like_types

    assert 'value' in query[field]
    value = query[field].get('value')
    assert type(value) is str

    return True

def validateTerms(query, mapping):
    assert type(query) is dict
    assert len(query) == 1

    field = [k for k in query][0]
    field_type = getTypeOfFieldFromMapping(field, mapping)
    assert field_type is not None
    assert field_type in keyword_like_types

    value = query[field]

    assert type(value) is list and False not in [type(v) is str for v in value]

    return True

def validateMatch(query, mapping):
    assert type(query) is dict
    assert len(query) == 1

    field = [k for k in query][0]
    field_type = getTypeOfFieldFromMapping(field, mapping)
    assert field_type is not None
    assert field_type in keyword_like_types

    value = query[field]
    assert type(value) is str

    return True

def validateRange(query, mapping):
    assert type(query) is dict
    assert len(query) == 1

    field = [k for k in query][0]
    field_type = getTypeOfFieldFromMapping(field, mapping)
    assert field_type is not None
    assert field_type in numeric_field_types

    value = query[field]
    assert type(value) is dict
    assert len(value) > 0
    valid_bounds = {'gt', 'lt', 'gte', 'lte'}
    assert False not in [key in valid_bounds for key in value]
    assert False not in [type(value[key]) in {float, int} for key in value]

    return True

def validateFieldBoost(query, mapping):
    assert type(query) is dict
    
    allowed_fields = {'field', 'boost'}
    assert False not in [field in allowed_fields for field in query]

    assert 'field' in query

    field = query['field']
    field_type = getTypeOfFieldFromMapping(field, mapping)
    assert field_type in numeric_field_types
    
    boost_value = query.get('boost', 1)
    assert type(boost_value) in {float, int}

    return True
    



query_validators = {
    'bool': validateBool,
    'filter': validateFilter,
    'must': validateMust,
    'should': validateShould,
    'must_not': validateMustNot,
    'term': validateTerm,
    'terms': validateTerms,
    'match': validateMatch,
    'range': validateRange,
    'fieldBoost': validateFieldBoost,
}


# given a query and a mapping, validate that the query
# follows the rules
def validateQuery(query, mapping):
    assert type(query is dict)
    assert len(query) == 1

    op = [k for k in query][0]
    assert op in query_validators
    return query_validators[op](query[op], mapping)
